import java.awt.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.util.*;

import com.umlet.control.*;
import com.umlet.element.base.*;
import com.umlet.element.base.detail.*;


/**
 * @author Markus Muckhoff
 * Lehrgebiet Software Engineering
 * FernUniversit�t Hagen
 * @version 1.0
 */

public class RelationDeluxe extends com.umlet.element.base.Relation { 
  String beginQualifier; 
  String endQualifier; 
  String beginArrow; 
  String endArrow; 
  String beginMultiplicity; 
  String endMultiplicity; 
  String beginRole; 
  String endRole; 
  String lineType; 
	
  private float TextPos, TextDist, TextRotation;
  private boolean TextRotationAlign, TextSegment, TextAlign;
  private int SegmentNumber;
  private Color savedfillColor;
  private boolean fill;

  Vector<String> _strings; 
  private Vector<String> getStrings() { 
	if (_strings==null) { 
	  _strings=new Vector<String>(); 
	} 
	return _strings; 
  } 
  private void setStrings(Vector<String> v) { 
	_strings = v; 
  } 

  private float getLength(Point p1, Point p2)
  {
	return (float)(Math.sqrt( (p2.x - p1.x) * (p2.x - p1.x) + (p2.y - p1.y) * (p2.y - p1.y) ));
  }
  
  private Point getNormal(Point p1, Point p2)
  {
	return new Point(p2.y - p1.y, p1.x - p2.x);
  }
  
 private Point getCenterOfLine()
 { 
	int i;
	float Poly_length = 0.0f, templength;
	Point ret=new Point(); 
	
	if((!TextSegment) || ((this.getLinePoints().size()  - 1) <= SegmentNumber) || (SegmentNumber < 0))
	{
		//Normal part******************************
		for(i=1;i<this.getLinePoints().size();i++)
		{
			Point p1 = this.getLinePoints().elementAt(i - 1);
			Point p2 = this.getLinePoints().elementAt(i);
			Poly_length += getLength(p1, p2);
		}
		
		Poly_length *= TextPos;
		
		for(i=1;i<this.getLinePoints().size();i++)
		{
			Point p1 = this.getLinePoints().elementAt(i - 1);
			Point p2 = this.getLinePoints().elementAt(i);
			Point Normal = getNormal(p1, p2);
			templength = getLength(p1, p2);
			
			if(templength > Poly_length)
			{
				ret.x = p1.x + (int)( (float)(p2.x - p1.x) / templength * Poly_length);
				ret.y = p1.y + (int)( (float)(p2.y - p1.y) / templength * Poly_length);
				
				ret.x += (int)(Normal.x * TextDist / templength);
				ret.y += (int)(Normal.y * TextDist / templength);
				
				break;
			}
			else
				Poly_length -= templength;
		}
		//Normal part******************************
	}
	else
	{
		Point p1 = this.getLinePoints().elementAt(SegmentNumber);
		Point p2 = this.getLinePoints().elementAt(SegmentNumber+1);
		Point Normal = getNormal(p1, p2);
		templength = getLength(p1, p2);
		
		ret.x = p1.x + (int)( (float)(p2.x - p1.x) * TextPos);
		ret.y = p1.y + (int)( (float)(p2.y - p1.y) * TextPos);
		
		ret.x += (int)(Normal.x * TextDist / templength);
		ret.y += (int)(Normal.y * TextDist / templength);
	}

	
	return ret; 
  }
  
  private double getSegmentAngle()
 { 
	int i;
	float Poly_length = 0.0f, templength;
	Point p1=new Point(), p2=new Point();
	
	 if((!TextSegment) || ((this.getLinePoints().size()  - 1) <= SegmentNumber) || (SegmentNumber < 0))
	{
		//Normal part******************************
		for(i=1;i<this.getLinePoints().size();i++)
		{
			p1 = this.getLinePoints().elementAt(i - 1);
			p2 = this.getLinePoints().elementAt(i);
			Poly_length += getLength(p1, p2);
		}
		
		Poly_length *= TextPos;
		
		for(i=1;i<this.getLinePoints().size();i++)
		{
			p1 = this.getLinePoints().elementAt(i - 1);
			p2 = this.getLinePoints().elementAt(i);
			//Point Normal = getNormal(p1, p2);
			templength = getLength(p1, p2);
			
			if(templength > Poly_length)
				break;
			else
				Poly_length -= templength;
		}
		//Normal part******************************
	}
	else
	{
		p1 = this.getLinePoints().elementAt(SegmentNumber);
		p2 = this.getLinePoints().elementAt(SegmentNumber+1);
	}
	
	return(Math.atan((double)(p2.y - p1.y) / (double)(p2.x - p1.x)));
  }
  
	private void setOutputTransformation(Graphics2D g2, Point p1, Point p2, float angle, float linePos, float lineDist)
	{
		float templength;
		Point ret=new Point();

		Point Normal = getNormal(p1, p2);
		templength = getLength(p1, p2);

		ret.x = p1.x + (int)( (float)(p2.x - p1.x) * linePos / templength);
		ret.y = p1.y + (int)( (float)(p2.y - p1.y) * linePos / templength);
		
		ret.x += (int)(Normal.x * lineDist / templength);
		ret.y += (int)(Normal.y * lineDist / templength);

		g2.translate(ret.x,ret.y);

		if(angle == 0.0f)
			angle = (float)Math.atan((double)(p2.y - p1.y) / (double)(p2.x - p1.x));
		g2.rotate(angle);
	}


  private void readTextPositionAttributes(String s)
  {
	String[] StringArray = s.split(";");
	
	for(int i=0;i<StringArray.length;i++)
	{
		if(StringArray[i].startsWith("p:") && (StringArray[i].length() > 2))
		{
			if( Float.valueOf(StringArray[i].substring(2, StringArray[i].length())) <= 1.0f )
				TextPos=Float.valueOf(StringArray[i].substring(2, StringArray[i].length()));
		}
		else if(StringArray[i].startsWith("d:") && (StringArray[i].length() > 2))
		{
			TextDist = Float.valueOf(StringArray[i].substring(2, StringArray[i].length()));
		}
		else if(StringArray[i].startsWith("r:") && (StringArray[i].length() > 2))
		{
			if( StringArray[i].substring(2, StringArray[i].length()).equals("*") )
				TextRotationAlign=true;
			else
				TextRotation = Float.valueOf(StringArray[i].substring(2, StringArray[i].length())) * (float)(Math.PI) / 180.0f;
		}
		else if(StringArray[i].startsWith("s:") && (StringArray[i].length() > 2))
		{
			TextSegment=true;
			SegmentNumber=Integer.valueOf(StringArray[i].substring(2, StringArray[i].length())) - 1;
		}
		else if(StringArray[i].startsWith("a:") && (StringArray[i].length() > 2))
		{
			if(StringArray[i].substring(2, StringArray[i].length()).equals("c"))
				TextAlign=true;
		}
	}
	
/*	
	if(StringArray.length > 0)
	{
		if( Float.valueOf(StringArray[0]) <= 1.0f )
			TextPos=Float.valueOf(StringArray[0]);
		
		if(StringArray.length > 1)
			TextDist = Float.valueOf(StringArray[1]);
		
		if(StringArray.length > 2)
			TextRotation = Float.valueOf(StringArray[2]) * (float)(Math.PI) / 180.0f;
	}
*/
  }

  public String getAdditionalAttributes() { 
	Vector<String> tmp=new Vector<String>(); 
	//tmp.add(beginQualifier); 
	//tmp.add(endQualifier); 
	//tmp.add(beginArrow); 
	//tmp.add(endArrow); 
	//tmp.add(beginMultiplicity); 
	//tmp.add(endMultiplicity); 
	//tmp.add(beginRole); 
	//tmp.add(endRole); 
	//tmp.add(lineType); 

	/*tmp.add(""+this.getX()); 
	tmp.add(""+this.getY()); 
	tmp.add(""+this.getWidth()); 
	tmp.add(""+this.getHeight());*/ 

	for (int i=0; i<this.getLinePoints().size(); i++) { 
	  Point p=this.getLinePoints().elementAt(i); 
	  String s1=""+p.x; 
	  String s2=""+p.y; 
	  tmp.add(s1); 
	  tmp.add(s2); 
	} 

	String ret=Constants.composeStrings(tmp,  Constants.DELIMITER_ADDITIONAL_ATTRIBUTES); 
	return ret; 
  } 
  public void setHiddenState(String s) { 
	Vector<String> tmp=Constants.decomposeStringsIncludingEmptyStrings(s,  Constants.DELIMITER_FIELDS); 
	// 9 attributes, 4 for start/end point, 4 for x,y,w,h 
	if (tmp.size()<17) throw new RuntimeException("UMLet: Error in  Relation.setHiddenState(), state value = "+s); 

	beginQualifier=tmp.elementAt(0); 
	endQualifier=tmp.elementAt(1); 
	beginArrow=tmp.elementAt(2); 
	endArrow=tmp.elementAt(3); 
	beginMultiplicity=tmp.elementAt(4); 
	endMultiplicity=tmp.elementAt(5); 
	beginRole=tmp.elementAt(6); 
	endRole=tmp.elementAt(7); 
	lineType=tmp.elementAt(8); 

	int x=Integer.parseInt(tmp.elementAt(9)); 
	int y=Integer.parseInt(tmp.elementAt(10)); 
	int w=Integer.parseInt(tmp.elementAt(11)); 
	int h=Integer.parseInt(tmp.elementAt(12)); 
	this.setBounds(x,y,w,h); 
	//this.setBounds(x-100,y-100,w+100,h+100);
	
	for (int i=13; i<tmp.size(); i=i+2) { 
	  int xx=Integer.parseInt(tmp.elementAt(i)); 
	  int yy=Integer.parseInt(tmp.elementAt(i+1)); 
	  Point p=new Point(xx,yy); 
	  this.getLinePoints().add(p); 
	} 
  } 

public void setAdditionalAttributes(String s) { 
	Vector<String> tmp=Constants.decomposeStringsIncludingEmptyStrings(s,  Constants.DELIMITER_ADDITIONAL_ATTRIBUTES); 
	// 9 attributes, 4 for start/end point, 4 for x,y,w,h 
	/*if (tmp.size()<17) throw new RuntimeException("UMLet: Error in  Relation.setHiddenState(), state value = "+s); 

	beginQualifier=(String)tmp.elementAt(0); 
	endQualifier=(String)tmp.elementAt(1); 
	beginArrow=(String)tmp.elementAt(2); 
	endArrow=(String)tmp.elementAt(3); 
	beginMultiplicity=(String)tmp.elementAt(4); 
	endMultiplicity=(String)tmp.elementAt(5); 
	beginRole=(String)tmp.elementAt(6); 
	endRole=(String)tmp.elementAt(7); 
	lineType=(String)tmp.elementAt(8); 

	int x=Integer.parseInt((String)tmp.elementAt(9)); 
	int y=Integer.parseInt((String)tmp.elementAt(10)); 
	int w=Integer.parseInt((String)tmp.elementAt(11)); 
	int h=Integer.parseInt((String)tmp.elementAt(12)); 
	this.setBounds(x,y,w,h);*/ 

	for (int i=0; i<tmp.size(); i=i+2) { 
	  int xx=Integer.parseInt(tmp.elementAt(i)); 
	  int yy=Integer.parseInt(tmp.elementAt(i+1)); 
	  Point p=new Point(xx,yy); 
	  this.getLinePoints().add(p); 
	} 
  } 


  public void setState(String state) { 
	beginQualifier=""; 
	endQualifier=""; 
	beginArrow=""; 
	endArrow=""; 
	beginMultiplicity=""; 
	endMultiplicity=""; 
	beginRole=""; 
	endRole=""; 
	lineType="-"; 
	this.setStrings(null); 

	_state=state; 

	Vector<String> tmp=Constants.decomposeStrings(state, "\n"); 

	//set the default for the TextPos
	TextPos=0.5f;
	TextDist=1.0f;
	TextRotation=0.0f;
	TextRotationAlign=false;
	TextSegment=false;
	TextAlign=false;
	//_fillColor=savedfillColor;
	fill=false;
	
	for (int i=0; i<tmp.size(); i++) { 
	  String s=tmp.elementAt(i); 
	  if (s.startsWith("q1=") & s.length()>3) { 
		beginQualifier=s.substring(3, s.length()); 
	  } else if (s.startsWith("q2=") & s.length()>3) { 
		endQualifier=s.substring(3, s.length()); 
	  } else if (s.startsWith("m1=") & s.length()>3) { 
		beginMultiplicity=s.substring(3, s.length()); 
	  } else if (s.startsWith("m2=") & s.length()>3) { 
		endMultiplicity=s.substring(3, s.length()); 
	  } else if (s.startsWith("r1=") & s.length()>3) { 
		beginRole=s.substring(3, s.length()); 
	  } else if (s.startsWith("r2=") & s.length()>3) { 
		endRole=s.substring(3, s.length()); 
	  } else if (s.startsWith("textpos=") & s.length()>8) {
		  readTextPositionAttributes(s.substring(8, s.length())); 		  
	  } else if (s.startsWith("lt=") & s.length()>3) { 
		if (s.indexOf("<<<<<")>=0) {
		  //_fillColor = _activeColor;
		  fill=true;
		  beginArrow="<<";
		} else if (s.indexOf("<<<<")>=0) { 
		  beginArrow="X"; 
		} else if (s.indexOf("<<<")>=0) { 
		  beginArrow="x"; 
		} else if (s.indexOf("<<")>=0) { 
		  beginArrow="<<"; 
		} else if (s.indexOf("<")>=0) { 
		  beginArrow="<"; 
		}
		if (s.indexOf(">>>>>")>=0) {
		  //_fillColor = _activeColor;
		  fill=true;
		  endArrow="<<";
		} else if (s.indexOf(">>>>")>=0) { 
		  endArrow="X"; 
		} else if (s.indexOf(">>>")>=0) { 
		  endArrow="x"; 
		} else if (s.indexOf(">>")>=0) { 
		  endArrow="<<"; 
		} else if (s.indexOf(">")>=0) { 
		  endArrow="<"; 
		} 
		if (s.indexOf(".")>=0) { 
		  lineType="."; 
		} 
	  } else { 
		this.getStrings().add(s); 
	  } 
	} 
  } 


  // Created objects have no sideeffects 
  // Only exception: no point is outside shape 
  // At least 2 points must be provided 
  public static Vector<Point> getIntersectingLineSegment(Area r, Vector<Point>  points) { 
	Vector<Point> ret=new Vector<Point>(); 

	// If no segment found, take last two points 
	Point pp_end=points.elementAt(points.size()-1); 
	Point pp_start=points.elementAt(points.size()-2); 

	for (int i=1;i<points.size();i++) { 
	  pp_end=points.elementAt(i); 
	  if (!r.contains(pp_end)) { 
		// End point of intersecting line found 
		pp_start=points.elementAt(i-1); 

		ret.add(pp_start); ret.add(pp_end); 
		return ret; 
	  } 
	} 

	ret.add(pp_start); ret.add(pp_end); 
	return ret; 
  } 

  public static Point moveNextTo(Area rFixed, Rectangle rMovable,  Point pStart, Point pEnd) { 
	// These ints can simply be added to line 
	int centerDiffX=(int)-rMovable.getWidth()/2; 
	int centerDiffY=(int)-rMovable.getHeight()/2; 

	int vectorX=pEnd.x-pStart.x; 
	int vectorY=pEnd.y-pStart.y; 

	int startx=pStart.x; 
	int starty=pStart.y; 
	int endx=pEnd.x; 
	int endy=pEnd.y; 

	for (int i=0;;i++) { 
	  endx+=vectorX; 
	  endy+=vectorY; 
	  rMovable.setLocation(endx+centerDiffX,endy+centerDiffY); 
	  if (!rFixed.intersects(rMovable)) break; 
	} 

	int newx=0; 
	int newy=0; 
	for (int i=0;i<10;i++) { 
	  newx=(endx+startx)/2; 
	  newy=(endy+starty)/2; 
	  rMovable.setLocation(newx+centerDiffX,newy+centerDiffY); 
	  if (rFixed.intersects(rMovable)) { 
		startx=newx; 
		starty=newy; 
	  } else { 
		endx=newx; 
		endy=newy; 
	  } 
	} 

	Point ret=new Point(newx+centerDiffX, newy+centerDiffY); 
	return ret; 
  } 

  public static boolean lineUp(Vector<Rectangle> shapes, Vector<Point> points, int  hotspotx, int hotspoty) { 
	// Remove point with the same coordinates 
	for (int i=points.size()-1;i>0;i--) { 
	  Point p1=points.elementAt(i); 
	  Point p2=points.elementAt(i-1); 
	  if (p1.x==p2.x & p1.y==p2.y) points.removeElementAt(i); 
	} 
	if (points.size()<=1) return false; 

	if (shapes.size()<=1) return true; 

	// Vector ret=new Vector(); 

	//Rectangle rFixed; 
	Rectangle rMovable; 
	Area tmpArea=new Area(); 
	for (int i=0; i<shapes.size()-1; i++) { 
	  Rectangle r=shapes.elementAt(i); 
	  if (i==0) { // The hotspot of the first element is set 
		Point p=points.elementAt(0); 
		r.setLocation(p.x-hotspotx,p.y-hotspoty); 
	  } 
	  Area a=new Area(r); 
	  tmpArea.add(a); 

	  //rFixed=(Rectangle)shapes.elementAt(i); 
	  rMovable=shapes.elementAt(i+1); 

	  /*if (i==0) { // The hotspot of the first element is set 
		Point p=(Point)points.elementAt(0); 
		rFixed.setLocation(p.x-hotspotx,p.y-hotspoty); 
	  }*/ 

	  Vector<Point> tmp=getIntersectingLineSegment(tmpArea, points); 
	  Point startIntersectingLine=tmp.elementAt(0); 
	  Point endIntersectingLine=tmp.elementAt(1); 

	  Point res=moveNextTo(tmpArea, rMovable, startIntersectingLine,  endIntersectingLine); 
	  //ret.add(res); 

	  if (rMovable instanceof Arrow) { 
		Arrow arrow=(Arrow)rMovable; 

		Point diffA=new  Point(-startIntersectingLine.x+endIntersectingLine.x,  -startIntersectingLine.y+endIntersectingLine.y); 
		Point diffB1=new Point(diffA.y, -diffA.x); 
		Point diffB2=new Point(-diffB1.x, -diffB1.y); 
		Point a1=new Point(2*diffA.x+diffB1.x, 2*diffA.y+diffB1.y); 
		Point a2=new Point(2*diffA.x+diffB2.x, 2*diffA.y+diffB2.y); 

		a1=Constants.normalize(a1,Constants.getFontsize()); 
		a2=Constants.normalize(a2,Constants.getFontsize()); 

		arrow.setArrowEndA(a1); 
		arrow.setArrowEndB(a2); 
	  } 

	  // ATTENTION: this Recangle will become the rFixed in the next  loop 
	  rMovable.setLocation(res); 
	} 

	return true; 
  } 

  public boolean isOnLine(int i) { 
	if (i-1>=0 & i+1<getLinePoints().size()) { 
	  Point x1=getLinePoints().elementAt(i-1); 
	  Point x2=getLinePoints().elementAt(i+1); 
	  Point p=getLinePoints().elementAt(i); 
	  if (pyth(p,x1)+pyth(p,x2)<pyth(x1,x2)+5) return true; 
	} 
	return false; 
  } 

  public int getWhereToInsert(Point p) { 
	for (int i=0; i<getLinePoints().size()-1;i++) { 
	  Point x1=getLinePoints().elementAt(i); 
	  Point x2=getLinePoints().elementAt(i+1); 
	  if (pyth(p,x1)+pyth(p,x2)<pyth(x1,x2)+5) return i+1; 
	} 
	return -1; 
  } 

  public int getLinePoint(Point p) { 
	for (int i=0; i<getLinePoints().size();i++) { 
	  Point x=getLinePoints().elementAt(i); 
	  if (pyth(p,x)<15) return i; 
	} 
	return -1; 
  } 

  private int pyth(Point x1, Point x2) { 
	int a=x1.x-x2.x; 
	int b=x1.y-x2.y; 
	return (int)Math.sqrt(a*a+b*b); 
  } 

  private <T> Vector<T> flipVector(Vector<T> v) { 
	Vector<T> ret=new Vector<T>(); 
	for (int i=v.size()-1; i>=0; i--) { 
	  ret.add(v.elementAt(i)); 
	} 
	return ret; 
  } 

  public boolean contains(Point p) { 
	for (int i=0; i<getLinePoints().size()-1;i++) { 
	  Point x1=getLinePoints().elementAt(i); 
	  Point x2=getLinePoints().elementAt(i+1); 
	  if (pyth(p,x1)+pyth(p,x2)<pyth(x1,x2)+5) return true; 
	} 
	for (int i=0; i<getLinePoints().size();i++) { 
	  Point x=getLinePoints().elementAt(i); 
	  if (pyth(p,x)<11) return true; 
	} 
	return false; 
  } 
  public boolean contains(int x, int y) { 
	return contains(new Point(x,y)); 
  } 

  private Vector<Point> _points; 
  public Vector<Point> getLinePoints() { 
	if (_points==null) { 
	  _points=new Vector<Point>(); 
	} 
	return _points; 
  } 
  public void setLinePoints(Vector<Point> v) { 
	_points=v; 
  } 

  public Entity CloneFromMe() { 
	RelationDeluxe c=new RelationDeluxe(); 

	c.setState(this.getPanelAttributes()); 
	c.setAdditionalAttributes(this.getAdditionalAttributes()); 

	c.setVisible(true); 
	c.setBounds(this.getBounds()); 
	return c; 
  } 

  public RelationDeluxe() { 
	super(); 
	savedfillColor = _fillColor;
  } 
  public RelationDeluxe(String s) { 
	super(s);
	savedfillColor = _fillColor;
  } 

  public void paint(Graphics g) { 
	Graphics2D g2=(Graphics2D) g; 
	g2.setFont(Constants.getFont()); 
	g2.setColor(_activeColor); 
	Constants.getFRC(g2); // Just to set anti-aliasing, even if no  text operations occur
	  
	_fillColor = (fill ? _activeColor : savedfillColor);
	  
	AffineTransform TempTransform;

	Vector<Rectangle> startShapes=new Vector<Rectangle>(); 
	Vector<Rectangle> endShapes=new Vector<Rectangle>(); 

	startShapes.add(new NoShape()); 
	endShapes.add(new NoShape()); 

	if (beginQualifier!=null && beginQualifier.length()>0) {
	Qualifier q;
		if(beginQualifier.equals("pin"))
		{
			q=new Qualifier("",  0,0,10,10); 
		}
		else if(beginQualifier.equals("bigpin"))
		{
			q=new Qualifier("",  0,0,20,20); 
		}
		else
		{
			TextLayout tl=new TextLayout(beginQualifier,  Constants.getFont(), Constants.getFRC(g2)); 
			q=new Qualifier(beginQualifier,  0,0,(int)tl.getBounds().getWidth()+Constants.getFontsize()*2,(int)tl.getBounds().getHeight()+Constants.getFontsize()/2); 
		}
		startShapes.add(q); 
	} 
	if (endQualifier!=null && endQualifier.length()>0) {
		Qualifier q;
		if(endQualifier.equals("pin"))
		{
			q=new Qualifier("",  0,0,10,10); 
		}
		else if(beginQualifier.equals("bigpin"))
		{
			q=new Qualifier("",  0,0,20,20); 
		}
		else
		{
			TextLayout tl=new TextLayout(endQualifier, Constants.getFont(),  Constants.getFRC(g2)); 
			q=new Qualifier(endQualifier,  0,0,(int)tl.getBounds().getWidth()+Constants.getFontsize()*2,(int)tl.getBounds().getHeight()+Constants.getFontsize()/2); 
		}
	  endShapes.add(q); 
	} 
	if (beginArrow!=null && beginArrow.length()>0) { 
	  Arrow a=new Arrow(beginArrow); 
	  startShapes.add(a); 
	} 
	if (endArrow!=null && endArrow.length()>0) { 
	  Arrow a=new Arrow(endArrow); 
	  endShapes.add(a); 
	} 
	if (beginMultiplicity!=null && beginMultiplicity.length()>0) { 
	  EmptyShape e=new EmptyShape(); 
	  startShapes.add(e); 
	  TextLayout tl=new TextLayout(beginMultiplicity,  Constants.getFont(), Constants.getFRC(g2)); 
	  Multiplicity m=new  Multiplicity(beginMultiplicity,0,0,(int)tl.getBounds().getWidth(),(int) tl.getBounds().getHeight()); 
	  startShapes.add(m); 
	} 
	if (endMultiplicity!=null && endMultiplicity.length()>0) { 
	  EmptyShape e=new EmptyShape(); 
	  endShapes.add(e); 
	  TextLayout tl=new TextLayout(endMultiplicity,  Constants.getFont(), Constants.getFRC(g2)); 
	  Multiplicity m=new  Multiplicity(endMultiplicity,0,0,(int)tl.getBounds().getWidth(),(int)tl .getBounds().getHeight()); 
	  endShapes.add(m); 
	} 
	if (beginRole!=null && beginRole.length()>0) { 
	  EmptyShape e=new EmptyShape(); 
	  startShapes.add(e); 
	  TextLayout tl=new TextLayout(beginRole, Constants.getFont(),  Constants.getFRC(g2)); 
	  Role r=new  Role(beginRole,0,0,(int)tl.getBounds().getWidth(),(int)tl.getBounds().getHeight()); 
	  startShapes.add(r); 
	} 
	if (endRole!=null && endRole.length()>0) { 
	  EmptyShape e=new EmptyShape(); 
	  endShapes.add(e); 
	  TextLayout tl=new TextLayout(endRole, Constants.getFont(),  Constants.getFRC(g2)); 
	  Role r=new  Role(endRole,0,0,(int)tl.getBounds().getWidth(),(int)tl.getBounds().getHeight()); 
	  endShapes.add(r); 
	} 

	Vector<Point> startPoints=new Vector<Point>(this.getLinePoints()); 
	Vector<Point> endPoints=flipVector(startPoints); 

	boolean a=lineUp(startShapes, startPoints, 0, 0); 
	boolean b=lineUp(endShapes, endPoints, 0, 0); 

	if (a==false || b==false) { 
	  return; 
	} 

	if (lineType.equals("-")) 
	  g2.setStroke(Constants.getStroke(0,1)); 
	else 
	  g2.setStroke(Constants.getStroke(1,1)); 
	for (int i=0; i<getLinePoints().size()-1; i++) { 
	  Point p1=getLinePoints().elementAt(i); 
	  Point p2=getLinePoints().elementAt(i+1); 
	  g2.drawLine(p1.x, p1.y, p2.x, p2.y); 
	} 
	g2.setStroke(Constants.getStroke(0,1)); 
	if (_selected) { 
	  for (int i=0; i<getLinePoints().size(); i++) { 
		Point p=getLinePoints().elementAt(i); 
		g2.drawOval(p.x-10, p.y-10, 20, 20); 
	  } 
	} 

	Vector<Rectangle> tmp=new Vector<Rectangle>(startShapes); 
	tmp.addAll(endShapes); 
	for (int i=0; i<tmp.size(); i++) { 
	  Rectangle r=tmp.elementAt(i); 
	  if (r instanceof Qualifier) { 
		Qualifier q=(Qualifier)r; 
//begin B. Buckl 
g.setColor(_fillColor); 
g.fillRect((int)r.getX(), (int)r.getY(), (int)r.getWidth(),  (int)r.getHeight()); 
g.setColor(_activeColor); 
//end 
		
		g.drawRect((int)r.getX(), (int)r.getY(), (int)r.getWidth(),  (int)r.getHeight()); 
		
		Constants.write(g2, q.getString(),  (int)r.getX()+Constants.getFontsize(),  (int)r.getY()+Constants.getFontsize(), false); 
	  } 
	  else if (r instanceof Arrow) { 
		Arrow arrow=(Arrow)r; 
		g2.drawLine((int)arrow.getX(), (int)arrow.getY(),  (int)arrow.getX()+(int)arrow.getArrowEndA().x,  (int)arrow.getY()+(int)arrow.getArrowEndA().y); 
		g2.drawLine((int)arrow.getX(), (int)arrow.getY(),  (int)arrow.getX()+(int)arrow.getArrowEndB().x,  (int)arrow.getY()+(int)arrow.getArrowEndB().y); 
		//System.out.println(arrow.getString()); 
		if (arrow.getString().equals("<<")) { 
	//begin B. Buckl 
	int[] ax=new int[3]; 
	int[] ay=new int[3]; 
	ax[0]=(int)arrow.getX(); 
	ax[1]=(int)arrow.getX()+(int)arrow.getArrowEndA().x; 
	ax[2]=(int)arrow.getX()+(int)arrow.getArrowEndB().x; 
	ay[0]=(int)arrow.getY(); 
	ay[1]=(int)arrow.getY()+(int)arrow.getArrowEndA().y; 
	ay[2]=(int)arrow.getY()+(int)arrow.getArrowEndB().y; 
	Polygon myPg = new Polygon(ax, ay, 3); 
	g2.setColor(_fillColor); 
	g2.fill(myPg); 
	g2.setColor(_activeColor); 
	g2.draw(myPg); 

	//g2.drawLine((int)arrow.getX()+(int)arrow.getArrowEndA().x,  (int)arrow.getY()+(int)arrow.getArrowEndA().y,  (int)arrow.getX()+(int)arrow.getArrowEndB().x,  (int)arrow.getY()+(int)arrow.getArrowEndB().y); 
} //end B. Buckl 
		else if (arrow.getString().equals("x")) { 
		  int[] ax=new int[4]; 
		  int[] ay=new int[4]; 
		  ax[0]=(int)arrow.getX(); 
		  ay[0]=(int)arrow.getY(); 
		  ax[1]=(int)arrow.getX()+(int)arrow.getArrowEndA().x; 
		  ay[1]=(int)arrow.getY()+(int)arrow.getArrowEndA().y; 
		  ax[3]=(int)arrow.getX()+(int)arrow.getArrowEndB().x; 
		  ay[3]=(int)arrow.getY()+(int)arrow.getArrowEndB().y; 

		  ax[2]=-(int)arrow.getX()+ax[1]+ax[3]; 
		  ay[2]=-(int)arrow.getY()+ay[1]+ay[3]; 

  //begin B. Buckl 
  Polygon myPg = new Polygon(ax, ay, 4); 
  g2.setColor(_fillColor); 
  g2.fill(myPg); 
  g2.setColor(_activeColor); 
  g2.draw(myPg); 
  //end B. Buckl 
		  } 
		else if (arrow.getString().equals("X")) { 
		  int[] ax=new int[4]; 
		  int[] ay=new int[4]; 
		  ax[0]=(int)arrow.getX(); 
		  ay[0]=(int)arrow.getY(); 
		  ax[1]=(int)arrow.getX()+(int)arrow.getArrowEndA().x; 
		  ay[1]=(int)arrow.getY()+(int)arrow.getArrowEndA().y; 
		  ax[3]=(int)arrow.getX()+(int)arrow.getArrowEndB().x; 
		  ay[3]=(int)arrow.getY()+(int)arrow.getArrowEndB().y; 

		  ax[2]=-(int)arrow.getX()+ax[1]+ax[3]; 
		  ay[2]=-(int)arrow.getY()+ay[1]+ay[3]; 

		  g2.fill(new Polygon(ax, ay, 4)); 
		  } 

	  } 
	  else if (r instanceof Multiplicity) { 
		Multiplicity m=(Multiplicity)r; 
		//g.drawRect((int)r.getX(), (int)r.getY(), (int)r.getWidth(),  (int)r.getHeight()); 

		if (m.getString() == endMultiplicity)
		{
			TempTransform = g2.getTransform();
			//r.setLocation((int)r.getX()+20, (int)r.getY());
			setOutputTransformation(g2, endPoints.elementAt(0), endPoints.elementAt(1), 0.0f, 30.0f, 0.0f);
			if(endPoints.elementAt(0).x > endPoints.elementAt(1).x)
				Constants.write(g2, m.getString(), -Constants.getPixelWidth(g2,m.getString()), -2,  false); //B.
			else
				Constants.write(g2, m.getString(), 0, -2,  false); //B.
			g2.setTransform(TempTransform);
		}

		if (m.getString() == beginMultiplicity)
		{
			TempTransform = g2.getTransform();
			//r.setLocation((int)r.getX()+20, (int)r.getY());
			setOutputTransformation(g2, startPoints.elementAt(0), startPoints.elementAt(1), 0.0f, 30.0f, 0.0f);
			if(startPoints.elementAt(0).x > startPoints.elementAt(1).x)
				Constants.write(g2, m.getString(), -Constants.getPixelWidth(g2,m.getString()), -2,  false); //B.
			else
				Constants.write(g2, m.getString(), 0, -2,  false); //B.
			g2.setTransform(TempTransform);
		}

		//Constants.write(g2, m.getString(), (int)r.getX(),  (int)r.getY()+Constants.getFontsize()+2*Constants.getDistTextToText(),  false); //B. Buckl added +2*Constants.getDistTextToText() 
	  }
	  else if (r instanceof Role) { 
		Role role=(Role)r; 

		if (role.getString() == endRole)
		{
			TempTransform = g2.getTransform();
			//r.setLocation((int)r.getX()+20, (int)r.getY());
			setOutputTransformation(g2, endPoints.elementAt(0), endPoints.elementAt(1), 0.0f, 20.0f, 0.0f);

			if(endPoints.elementAt(0).x > endPoints.elementAt(1).x)
				Constants.write(g2, role.getString(), -Constants.getPixelWidth(g2,role.getString()), Constants.getFontsize()+2*Constants.getDistTextToText(),  false);
			else
				Constants.write(g2, role.getString(), 0, Constants.getFontsize()+2*Constants.getDistTextToText(),  false);
			g2.setTransform(TempTransform);
		}


		if (role.getString() == beginRole)
		{
			TempTransform = g2.getTransform();
			//r.setLocation((int)r.getX()+20, (int)r.getY());
			setOutputTransformation(g2, startPoints.elementAt(0), startPoints.elementAt(1), 0.0f, 20.0f, 0.0f);
			if(startPoints.elementAt(0).x > startPoints.elementAt(1).x)
				Constants.write(g2, role.getString(), -Constants.getPixelWidth(g2,role.getString()), Constants.getFontsize()+2*Constants.getDistTextToText(),  false);
			else
				Constants.write(g2, role.getString(), 0, Constants.getFontsize()+2*Constants.getDistTextToText(),  false);
			g2.setTransform(TempTransform);
		}
		//else
		//g.drawRect((int)r.getX(), (int)r.getY(), (int)r.getWidth(),  (int)r.getHeight()); 
		//	Constants.write(g2, role.getString(), (int)r.getX(),  (int)r.getY()+Constants.getFontsize()+2*Constants.getDistTextToText(),  false); //B. Buckl added +2*Constants.getDistTextToText() 
	  } 
	} 

	if (this.getStrings()!=null) { 
	  if (this.getStrings().size()>0) { 
		Point start=this.getCenterOfLine(); 
		int yPos=start.y-Constants.getDistTextToText(); //B. Buckl  added -Constants.getDistTextToText()
		int xPos=start.x;
		int ytemp=0;
		
		TempTransform = g2.getTransform();
		//g2.translate(TempTransform.getTranslateX() + xPos, TempTransform.getTranslateX() + yPos);
		g2.translate(xPos,yPos);
		if(TextRotationAlign)
			TextRotation=(float)getSegmentAngle();
		g2.rotate(TextRotation);
		//xPos=0;
		//yPos=0;
		  
		for (int i=0; i<getStrings().size(); i++) { 
		  String s=this.getStrings().elementAt(i); 
		  //Constants.write(g2, s, xPos, yPos, TextAlign);
		  //yPos+=Constants.getFontsize(); 
		  //yPos+=Constants.getDistTextToText(); 
		  Constants.write(g2, s, 0, ytemp, TextAlign); 
		  ytemp+=Constants.getFontsize(); 
		  ytemp+=Constants.getDistTextToText(); 
		}
		
		//g2.rotate(-TextRotation);
		//g2.translate(-xPos,-yPos);
		g2.setTransform(TempTransform);
	  } 
	} 


	Vector<Point> criticalPoints=new Vector<Point>(); 
	for (int i=1; i<startShapes.size(); i++) { 
	  Rectangle r=startShapes.elementAt(i); 
	  Point p1=new Point((int)r.getX()-2, (int)r.getY()-2); 
	  Point p2=new Point((int)r.getX()+(int)r.getWidth()+2,  (int)r.getY()+(int)r.getHeight()+2); 
	  criticalPoints.add(p1); 
	  criticalPoints.add(p2); 
	} 
	for (int i=1; i<endShapes.size(); i++) { 
	  Rectangle r=endShapes.elementAt(i); 
	  Point p1=new Point((int)r.getX()-2, (int)r.getY()-2); 
	  Point p2=new Point((int)r.getX()+(int)r.getWidth()+2,  (int)r.getY()+(int)r.getHeight()+2); 
	  criticalPoints.add(p1); 
	  criticalPoints.add(p2); 
	} 
	if (this.getStrings()!=null) { 
	  if (this.getStrings().size()>0) { 
		Point start=this.getCenterOfLine(); 
		int yPos=start.y; 
		int xPos=start.x; 
		for (int i=0; i<getStrings().size(); i++) { 
		  String s=this.getStrings().elementAt(i); 
		  int width=Constants.getPixelWidth(g2, s); 
		  criticalPoints.add(new Point(xPos-width/2-20,  yPos-Constants.getFontsize()-20)); 
		  criticalPoints.add(new Point(xPos+width/2+20, yPos+20)); 
		  yPos+=Constants.getFontsize(); 
		  yPos+=Constants.getDistTextToText(); 
		} 
	  } 
	} 

	{ 
	  int minx=99999; int maxx=-99999; 
	  int miny=99999; int maxy=-99999; 
	  for (int i=0; i<getLinePoints().size(); i++) { 
		Point p=getLinePoints().elementAt(i); 
		minx=Math.min(minx,p.x); 
		miny=Math.min(miny,p.y); 
		maxx=Math.max(maxx,p.x); 
		maxy=Math.max(maxy,p.y); 

		//TEST 
		minx=Math.min(minx,p.x-20); 
		miny=Math.min(miny,p.y-20); 
		maxx=Math.max(maxx,p.x+20); 
		maxy=Math.max(maxy,p.y+20); 
	  } 
	  for (int i=0; i<criticalPoints.size(); i++) { 
		Point p=criticalPoints.elementAt(i); 
		minx=Math.min(minx,p.x); 
		miny=Math.min(miny,p.y); 
		maxx=Math.max(maxx,p.x); 
		maxy=Math.max(maxy,p.y); 
	  } 

	  int diffminx=minx; 
	  int diffminy=miny; 
	  int diffmaxx=maxx-getWidth(); 
	  int diffmaxy=maxy-getHeight(); 

	  if (minx!=0) { 
		Controller.getInstance().executeCommandWithoutUndo(new  Resize(this,8,diffminx,0)); 
		} 
	  if (maxx!=0) { 
		Controller.getInstance().executeCommandWithoutUndo(new  Resize(this,2,diffmaxx,0)); 
		} 
	  if (miny!=0) { 
		Controller.getInstance().executeCommandWithoutUndo(new  Resize(this,1,0,diffminy)); 
		} 
	  if (maxy!=0) { 
		Controller.getInstance().executeCommandWithoutUndo(new  Resize(this,4,0,diffmaxy)); 
	  } 
	  if (minx!=0 | miny!=0) { 
		for (int i=0; i<getLinePoints().size(); i++) { 
		  Point p=getLinePoints().elementAt(i); 
		  p.x+=-minx; 
		  p.y+=-miny; 
		} 
	  } 
	} 
  } 

  private Point getStartPoint() { 
	Point ret=this.getLinePoints().elementAt(0); 
	return ret; 
  } 
  private Point getEndPoint() { 
	Point  ret=this.getLinePoints().elementAt(this.getLinePoints().size() -1); 
	return ret; 
  } 

  public Point getAbsoluteCoorStart() { 
	Point ret=new Point(); 
	ret.x=this.getX()+this.getStartPoint().x; 
	ret.y=this.getY()+this.getStartPoint().y; 
	return ret; 
  } 
  public Point getAbsoluteCoorEnd() { 
	Point ret=new Point(); 
	ret.x=this.getX()+this.getEndPoint().x; 
	ret.y=this.getY()+this.getEndPoint().y; 
	return ret; 
  } 

  public int getConnectedLinePoint(Entity e, int action) { 
	int ret1=-1; 
	int ret2=-1; 

  Point p=this.getStartPoint(); 
  Point abs=new Point((int)(this.getX()+p.getX()),  (int)(this.getY()+p.getY())); 
  int where=e.doesCoordinateAppearToBeConnectedToMe(abs); 
  if ((where>0) && ((where & action)>0)) ret1=0; 

  p=this.getEndPoint(); 
  abs=new Point((int)(this.getX()+p.getX()),  (int)(this.getY()+p.getY())); 
  where=e.doesCoordinateAppearToBeConnectedToMe(abs); 
  if ((where>0) && ((where & action)>0))  ret2=this.getLinePoints().size()-1; 

	ret1++; 
	ret2++; 

	return ret1+1000*ret2; 
  } 
  
  public StickingPolygon getStickingBorder() { //LME
    return null;
  }
} 
